
public const string NOTIFY_STYLE = """
* {
    /*-notify-shadow: 0px 2px 18px alpha (#000, 0.60);*/
    
    border-radius: 4px;
    border-width: 1px;
    border-style: solid;
    
    border-color: alpha (#000, 0.3);
    border-top-color: #fbfbfb;
    
    background-color: alpha (#f9f9f9, 0.9);
}

.low {
    background-color: alpha (#f9f9f9, 0.5);
}

.critical {
    background-color: alpha (#f50000, 0.9);
    border-top-color: #f00;
}
""";

public const int WIDTH = 300;
public const int HEIGHT = 76; //56+2*10px shadow
public const int SHADOW = 10;
public const int EXPIRE = 7000; //in ms

public class Notification : Clutter.Group {
    
    public enum Urgency {
        LOW = 0,
        NORMAL = 1,
        CRITICAL = 2
    }
    
    Granite.Drawing.BufferSurface buffer;
    public Clutter.Texture closeb;
    
    public NotifyServer server;
    public Urgency urgency;
    
    public bool block_anim;
    
    public uint32 id;
    
    public GtkClutter.Texture tex;
    public Clutter.Text text;
    
    public Notification (string name, string body, string icon, uint32 timeout, string app, NotifyServer server, HashTable<string,Variant> hints, uint32 id) {
        this.server = server;
        
        this.id = id;
        this.block_anim = true;
        
        //create the widget used for drawing
        try {
            this.closeb = new Clutter.Texture.from_file (Constants.PKGDATADIR+"/close.png");
        } catch (Error e) { warning (e.message); }
        var draw_ref = new Gtk.Window ();
        var draw = new Clutter.CairoTexture (WIDTH, HEIGHT);
        
        
        //urgency
        if (hints.contains ("urgency")) {
            this.urgency = (Urgency)hints.lookup ("urgency").get_byte ();
            if (this.urgency == Urgency.CRITICAL)
                draw_ref.get_style_context ().add_class ("critical");
            else if (this.urgency == Urgency.LOW)
                draw_ref.get_style_context ().add_class ("low");
        } else {
            this.urgency = Urgency.NORMAL;
        }
        
        var css = new Gtk.CssProvider ();
        try {
            css.load_from_data (NOTIFY_STYLE, -1);
        } catch (Error e) { warning (e.message); }
        draw_ref.get_style_context ().add_provider (css, 20000);
        
        this.allocation_changed.connect ( () => {
            this.buffer = new Granite.Drawing.BufferSurface ((int)width, (int)height);
            Granite.Drawing.Utilities.cairo_rounded_rectangle (this.buffer.context, SHADOW, SHADOW, 
                this.width-2*SHADOW, this.height-2*SHADOW, 7);
            this.buffer.context.set_source_rgba (0, 0, 0, 0.4);
            this.buffer.context.fill ();
            this.buffer.exponential_blur (SHADOW/2);
            draw_ref.get_style_context ().render_activity (this.buffer.context, SHADOW, SHADOW, 
                this.width-2*SHADOW, this.height-2*SHADOW);
        });
        draw.draw.connect ( (ctx) => {
            ctx.set_source_surface (this.buffer.surface, 0, 0);
            ctx.paint ();
            return false;
        });
        draw.auto_resize = true;
        draw.add_constraint (new Clutter.BindConstraint (this, Clutter.BindCoordinate.WIDTH, 0));
        draw.add_constraint (new Clutter.BindConstraint (this, Clutter.BindCoordinate.HEIGHT, 0));
        
        
        tex = new GtkClutter.Texture ();
        this.set_pix (hints, app, icon);
        
        /*tex.width  = 48;
        tex.height = 48;*/
        
        
        
        text = new Clutter.Text.with_text ("", "");
        
        this.set_text (name, body);
        text.x = 50;
        text.add_constraint (new Clutter.BindConstraint (draw, Clutter.BindCoordinate.HEIGHT, -SHADOW*2-5));
        text.width = WIDTH - (text.x+SHADOW);
        text.line_wrap = true;
        text.line_wrap_mode = Pango.WrapMode.WORD_CHAR;
        text.ellipsize = Pango.EllipsizeMode.END;
        
        this.add_child (draw);
        this.add_child (tex);
        this.add_child (text);
        this.add_child (closeb);
        
        closeb.reactive = true;
        closeb.width = 25;
        closeb.height = 25;
        closeb.scale_gravity = Clutter.Gravity.CENTER;
        closeb.scale_x = 0;
        closeb.scale_y = 0;
        
        text.y = SHADOW+8;
        tex.y  = SHADOW+8;
        tex.x  = SHADOW+5;
        
        this.width  = WIDTH;
        this.reactive = true;
        
        for (var i=0;i<hints.get_keys ().length ();i++)
            print ("%s, ", hints.get_keys ().nth_data (i));
        print ("\n");
        
        uint timer = 0;
        
        //slide in anim
        this.rotation_angle_x = 90.0f;
        this.opacity = 0;
        this.height = 25.0f;
        this.rotation_center_x = {this.width/2, this.height/2, -100};
        this.animate (Clutter.AnimationMode.EASE_OUT_BOUNCE, 650, rotation_angle_x:0.0f, opacity:255, height:0.0f+HEIGHT).
            completed.connect ( () => {
            this.block_anim = false;
            if (this.urgency != Urgency.CRITICAL) {
                timer = Timeout.add ((timeout == uint32.MAX)?EXPIRE:timeout, () => {
                    this.close ();
                    return false;
                });
            }
        });
        
        //hover
        this.enter_event.connect ( () => {
            if (this.block_anim)
                return false;
            closeb.animate (Clutter.AnimationMode.EASE_OUT_ELASTIC, 400, scale_x:1.0f, scale_y:1.0f);
            
            if (this.urgency != Urgency.CRITICAL) // Stop the timer
                Source.remove (timer);
            
            float height;
            text.get_preferred_size (null, null, null, out height);
            var new_h = this.height + (height - text.height) + SHADOW;
            if (new_h < HEIGHT)
                new_h = HEIGHT;
            this.animate (Clutter.AnimationMode.EASE_OUT_QUAD, 400, height:new_h);
            return true;
        });
        this.leave_event.connect ( () => {
            if (this.block_anim)
                return false;
            closeb.animate (Clutter.AnimationMode.EASE_IN_QUAD, 400, scale_x:0.0f, scale_y:0.0f);
            
            // Reset the timer for 3 seconds
            if (this.urgency != Urgency.CRITICAL) {
                timer = Timeout.add ((timeout == uint32.MAX) ? 3000 : timeout, () => {
                    this.close ();
                    return false;
                });
            }
            
            this.animate (Clutter.AnimationMode.EASE_OUT_QUAD, 400, height:HEIGHT+0.0f);
            return true;
        });
        
        //close notification
        closeb.button_press_event.connect ( () => {
            this.close (true);
            return true;
        });
    }
    
    public void set_pix (HashTable<string, Variant> hints, string app, string icon) {
        /*decide on the icon, order:
         image-data
         image-path
         app_icon
         icon_data
         from app name?
         fallback to dialog-information
        */
        if (hints.contains ("image-data")) {                        /*image-data*/
            debug ("Using image-data");
            
            Gdk.Pixdata data = {};
            try {
                if (data.deserialize ((uint8[])hints.lookup ("image_data").dup_bytestring_array ()))
                    tex.set_from_pixbuf (Gdk.Pixbuf.from_pixdata (data));
                else
                    warning ("Error while deserializing image-data");
            } catch (Error e) { warning (e.message); }
        } else if (hints.contains ("image-path")) {                 /*image-path*/
            debug ("Using image-path");
            if (hints.lookup ("image-path").get_string ().has_prefix ("file://")) {
                var file = File.new_for_uri (hints.lookup ("image-path").get_string ());
                tex.filename = file.get_path ();
            } else {
                try {
                    tex.set_from_icon_name (new Gtk.Image (), 
                        hints.lookup ("image-path").get_string (), Gtk.IconSize.DND);
                } catch (Error e) { warning (e.message); }
            }
        } else if (icon != "") {                                    /*app_icon*/
            debug ("Using app icon");
            try {
                tex.set_from_icon_name (new Gtk.Image (), icon, Gtk.IconSize.DND);
            } catch (Error e) {  //not an icon name? try a stock icon
                try {
                    print ("Trying stock icon %s\n", icon);
                    tex.set_from_stock (new Gtk.Image (), icon, Gtk.IconSize.DND);
                } catch (Error e) { //still not? well... the app name perhaps?
                    try {
                        print ("Trying app icon %s\n", app);
                        tex.set_from_icon_name (new Gtk.Image (), app, Gtk.IconSize.DND);
                    } catch (Error e) { warning (e.message); } //hm...
                }
            }
        } else if (hints.contains ("icon_data")) {                  /*icon_data*/
            debug ("Using icon_data");
            Gdk.Pixdata data = {};
            try {
                if (data.deserialize ((uint8[])hints.lookup ("image-data").get_data ()))
                    tex.set_from_pixbuf (Gdk.Pixbuf.from_pixdata (data));
                else
                    warning ("Error while deserializing icon_data");
            } catch (Error e) { warning (e.message); }
        } else {
            try {                                                   /*icon from the app perhaps?*/
                debug ("Using app name");
                tex.set_from_icon_name (new Gtk.Image (), app.down (), Gtk.IconSize.DND);
            } catch (Error e) {
                debug ("Using fallback");
                try {                                               /*dialog-information icon as fallback*/
                    tex.set_from_icon_name (new Gtk.Image (), "dialog-information", Gtk.IconSize.DND);
                } catch (Error e) { warning (e.message); }
            }
        }
    }
    
    public void set_text (string name, string body) {
        text.text = "<span font='10' weight='bold'>"+name+"</span>\n<span font='9'>"+body+"</span>";
        text.use_markup = true;
    }
    
    public void close (bool fast_close=false) {
        this.block_anim = true;
        Clutter.Animation a;
        if (!fast_close) {
            a = this.animate (Clutter.AnimationMode.EASE_OUT_QUAD, 1000, opacity:0);
        } else {
            this.scale_gravity = Clutter.Gravity.CENTER;
            a = this.animate (Clutter.AnimationMode.EASE_IN_BOUNCE, 600, scale_x:0.0f, scale_y:0.0f);
        }
        a.completed.connect ( () => {
            if (this.get_parent () != null)
                this.get_parent ().remove_child (this);
            
            this.dispose ();
            
            server.update_input_shape ();
        });
    }
    
    
}

[DBus (name="org.freedesktop.Notifications")]
public class NotifyServer : Object {
    
    internal int id_counter = 0;
    
    public Granite.Widgets.CompositedWindow main_window;
    public Clutter.Box box;
    
    public NotifyServer () {
        var clutter = new GtkClutter.Embed ();
        
        this.main_window = new Granite.Widgets.CompositedWindow ();
        this.main_window.type_hint = Gdk.WindowTypeHint.NOTIFICATION;
        this.main_window.move (Gdk.Screen.get_default ().width ()-100, 0);
        this.main_window.stick ();
        this.main_window.set_keep_above (true);
        this.main_window.focus_on_map = false;
        this.main_window.accept_focus = false;
        this.main_window.skip_pager_hint = true;
        this.main_window.skip_taskbar_hint = true;
        this.main_window.focus_visible = false;
        this.box = new Clutter.Box (new Clutter.BoxLayout ());
        
        ((Clutter.BoxLayout)this.box.layout_manager).vertical = true;
        this.box.add_constraint (new Clutter.BindConstraint (clutter.get_stage (), Clutter.BindCoordinate.WIDTH, 0));
        this.box.add_constraint (new Clutter.BindConstraint (clutter.get_stage (), Clutter.BindCoordinate.HEIGHT, 0));
        
        ((Clutter.Stage)clutter.get_stage ()).color = {0, 0, 0, 0};
        ((Clutter.Stage)clutter.get_stage ()).use_alpha = true;
        clutter.get_stage ().add_child (box);
        
        this.main_window.add (clutter);
        this.main_window.set_size_request (WIDTH, Gdk.Screen.get_default ().height ());
        this.main_window.show_all ();
        
        this.update_input_shape ();
    }
    
    public void update_input_shape () {
        this.main_window.get_window ().input_shape_combine_region (new Cairo.Region.rectangle 
            ({0, 0, WIDTH, this.box.get_n_children () * HEIGHT}), 0, 0);
    }
    
    public void close_notification (uint32 id) {
        
    }
    public string [] get_capabilities () {
        return {"body", "body-markup", null};
    }
    public void get_server_information (out string name, out string vendor, out string version, out string spec_version) {
        name = "pantheon-notify";
        vendor = "ElementaryOS";
        version = "0.1";
        spec_version = "1.1";
    }
    public new uint32 notify (string app_name, uint32 replaces_id, string app_icon, string summary, 
        string body, string [] actions, HashTable<string, Variant> hints, int32 expire_timeout) {
        
        
        int urgency = 1;
        if (hints.contains ("urgency"))
            urgency = hints.lookup ("urgency").get_byte ();
        print ("New notification, ID: %i, app: %s, icon: %s, priority: %i\n", id_counter + 1, 
                app_name, app_icon, urgency);
        
        
        if (replaces_id != 0) { //replace the notification
            
            bool found_one = false;
            for (var i=0;i<this.box.get_children ().length ();i++) {
                if (((Notification)this.box.get_children ().nth_data (i)).id == replaces_id) {
                    
                    ((Notification)this.box.get_children ().nth_data (i)).set_text (summary, body);
                    ((Notification)this.box.get_children ().nth_data (i)).set_pix  (hints, app_name, app_icon);
                    
                    found_one = true;
                    break;
                }
            }
            
            if (!found_one) //just add a normal one
                this.box.insert_child_at_index (new Notification (summary, body, app_icon, 
                    expire_timeout, app_name, this, hints, replaces_id), 0);
            
            this.update_input_shape ();
            
            return replaces_id;
        } else {
            
            this.box.insert_child_at_index (new Notification (summary, body, app_icon, 
                expire_timeout, app_name, this, hints, id_counter+1), 0);
            
            this.update_input_shape ();
            return id_counter ++;
        }
    }
}

public static void main (string [] args) {
    GtkClutter.init (ref args);
    
    Granite.Services.Logger.initialize ("pantheon-notifications");
    Granite.Services.Logger.DisplayLevel = Granite.Services.LogLevel.DEBUG;
    
    var n = new NotifyServer ();
    Bus.own_name (
        BusType.SESSION, 
        "org.freedesktop.Notifications", 
        BusNameOwnerFlags.NONE, 
        (con) => {
            try {
                con.register_object ("/org/freedesktop/Notifications", n);
            } catch (Error e) { warning (e.message); }
        }, 
        () => {}, 
        (con, name)=> warning ("Could not aquire bus %s", name));
    
    Gtk.main ();
}

